﻿/*
 * Copyright 2020 faddenSoft
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
using System;
using System.Collections.ObjectModel;

using PluginCommon;

namespace Freitag {
    /// <summary>
    /// Converts monchrome bitmaps intended for the CHARSET function in GRAPHIX II.
    ///
    /// (I don't know why it's called "CHARSET".)
    /// </summary>
    public class VisCharset : MarshalByRefObject, IPlugin, IPlugin_Visualizer {
        // IPlugin
        public string Identifier {
            get { return "Caverns of Freitag CHARSET Visualizer"; }
        }
        private IApplication mAppRef;
        private byte[] mFileData;

        // Visualization identifiers; DO NOT change or projects that use them will break.
        private const string VIS_GEN_IMAGE_SET = "freitag-charset-set";
        private const string VIS_GEN_IMAGE = "freitag-charset-single";

        private const string P_OFFSET = "offset";
        private const string P_INDEX = "index";
        private const string P_WIDTH = "width";
        private const string P_HEIGHT = "height";

        private const int MAX_DIM = 4096;

        // Visualization descriptors.
        private VisDescr[] mDescriptors = new VisDescr[] {
            new VisDescr(VIS_GEN_IMAGE_SET, "Freitag CHARSET Image Set", VisDescr.VisType.Bitmap,
                new VisParamDescr[] {
                    new VisParamDescr("File offset (hex)",
                        P_OFFSET, typeof(int), 0, 0x00ffffff, VisParamDescr.SpecialMode.Offset, 0),
                    new VisParamDescr("Image index",
                        P_INDEX, typeof(int), 0, 256, 0, 0),
                }),
            new VisDescr(VIS_GEN_IMAGE, "Freitag CHARSET Single Image", VisDescr.VisType.Bitmap,
                new VisParamDescr[] {
                    new VisParamDescr("File offset (hex)",
                        P_OFFSET, typeof(int), 0, 0x00ffffff, VisParamDescr.SpecialMode.Offset, 0),
                    new VisParamDescr("Width",
                        P_WIDTH, typeof(int), 1, 40, 0, 1),
                    new VisParamDescr("Height",
                        P_HEIGHT, typeof(int), 1, 192, 0, 1),
                })
        };


        // IPlugin
        public void Prepare(IApplication appRef, byte[] fileData, AddressTranslate addrTrans) {
            mAppRef = appRef;
            mFileData = fileData;
        }

        // IPlugin
        public void Unprepare() {
            mAppRef = null;
            mFileData = null;
        }

        // IPlugin_Visualizer
        public VisDescr[] GetVisGenDescrs() {
            if (mFileData == null) {
                return null;
            }
            return mDescriptors;
        }

        // IPlugin_Visualizer
        public IVisualization2d Generate2d(VisDescr descr,
                ReadOnlyDictionary<string, object> parms) {
            switch (descr.Ident) {
                case VIS_GEN_IMAGE_SET:
                    return GenerateImageFromSet(parms);
                case VIS_GEN_IMAGE:
                    return GenerateImage(parms);
                default:
                    mAppRef.ReportError("Unknown ident " + descr.Ident);
                    return null;
            }
        }

        private IVisualization2d GenerateImageFromSet(ReadOnlyDictionary<string, object> parms) {
            int offset = Util.GetFromObjDict(parms, P_OFFSET, 0);
            int imageIndex = Util.GetFromObjDict(parms, P_INDEX, 0);

            if (offset < 0 || offset >= mFileData.Length || imageIndex < 0 || imageIndex > 255) {
                mAppRef.ReportError("Invalid parameter");
                return null;
            }

            int maxIndex = mFileData[offset];   // count + 1
            if (imageIndex > maxIndex) {
                mAppRef.ReportError("Image index " + imageIndex +
                    " exceeds table size " + maxIndex);
                return null;
            }

            int hdrOffset = offset + 1 + imageIndex * 4;
            int lastOffset = hdrOffset + 3;
            if (lastOffset >= mFileData.Length) {
                mAppRef.ReportError("Header would be off end of file (last offset +" +
                    lastOffset.ToString("x6") + ")");
                return null;
            }

            int itemOffset = mFileData[hdrOffset] | (mFileData[hdrOffset + 1] << 8);
            int width = mFileData[hdrOffset + 2];
            int height = mFileData[hdrOffset + 3];

            if (width == 0 || height == 0) {
                mAppRef.ReportError("Item " + imageIndex + " has width and/or height of zero");
            }

            lastOffset = offset + itemOffset + ((width + 7) / 8) * height - 1;
            if (lastOffset >= mFileData.Length) {
                mAppRef.ReportError("Image runs off end of file (last offset +" +
                    lastOffset.ToString("x6") + ")");
                return null;
            }

            return GenerateBitmap(itemOffset, width, height);
        }

        private IVisualization2d GenerateImage(ReadOnlyDictionary<string, object> parms) {
            int offset = Util.GetFromObjDict(parms, P_OFFSET, 0);
            int width = Util.GetFromObjDict(parms, P_WIDTH, 1);
            int height = Util.GetFromObjDict(parms, P_HEIGHT, 1);

            if (offset < 0 || offset >= mFileData.Length ||
                    width <= 0 || width > MAX_DIM ||
                    height <= 0 || height > MAX_DIM) {
                mAppRef.ReportError("Invalid parameter");
                return null;
            }

            int lastOffset = offset + ((width + 7) / 8) * height - 1;
            if (lastOffset >= mFileData.Length) {
                mAppRef.ReportError("Image runs off end of file (last offset +" +
                    lastOffset.ToString("x6") + ")");
                return null;
            }

            return GenerateBitmap(offset, width, height);
        }

        private IVisualization2d GenerateBitmap(int dataOffset, int width, int height) {
            VisBitmap8 vb = new VisBitmap8(width, height);
            SetPalette(vb);

            int bit = 0x01;
            byte val = mFileData[dataOffset++];
            for (int row = 0; row < height; row++) {
                for (int col = 0; col < width; col++) {
                    if ((val & bit) != 0) {
                        vb.SetPixelIndex(col, row, 1);
                    }

                    bit <<= 1;
                    if (bit == 0x0100) {
                        bit = 0x01;
                        val = mFileData[dataOffset++];
                    }
                }
            }

            return vb;
        }

        private void SetPalette(VisBitmap8 vb) {
            vb.AddColor(0x00, 0x00, 0x00, 0x00);    // 0=transparent
            vb.AddColor(0xff, 0x20, 0x20, 0xe0);    // 1=solid (mostly blue)
        }
    }
}
